// Source file for the interactive laplacian editing program



////////////////////////////////////////////////////////////////////////
// Include files
////////////////////////////////////////////////////////////////////////

#include <vector>
using namespace std;
#include <GL/glut.h>
#include "R3Graphics/R3Graphics.h"
#include "R3LaplacianMesh.h"



////////////////////////////////////////////////////////////////////////
// Data type definitions
////////////////////////////////////////////////////////////////////////

struct Anchor {
  int vertex_id;
  R3Point position;
};



////////////////////////////////////////////////////////////////////////
// Program variables
////////////////////////////////////////////////////////////////////////

// Program arguments

static char *input_mesh_name = NULL;
static char *output_mesh_name = NULL;
static char *input_function_name = NULL;
static char *output_function_name = NULL;
static char *output_image_name = NULL;
static double empty_value = 0;
static int print_verbose = 0;
static int batch = 0;


// Data variables

static R3LaplacianMesh *mesh = NULL;
static R3Viewer *viewer = NULL;
static vector<Anchor *> anchors;
static Anchor *current_anchor = NULL;


// Display option variables

static int show_anchors = 1;
static int show_values = 1;
static int show_faces = 1;
static int show_edges = 0;
static int show_vertices = 0;
static int show_backfacing = 1;
static R3Point center(0,0,0);


// GLUT interface variables 

static int GLUTwindow = 0;
static int GLUTwindow_width = 800;
static int GLUTwindow_height = 600;
static int GLUTmouse[2] = { 0, 0 };
static int GLUTbutton[3] = { 0, 0, 0 };
static int GLUTmodifiers = 0;



////////////////////////////////////////////////////////////////////////
// Mesh Input/Output functions
////////////////////////////////////////////////////////////////////////

static R3LaplacianMesh *
ReadFiles(const char *mesh_name, const char *function_name)
{
  // Allocate mesh
  R3LaplacianMesh *mesh = new R3LaplacianMesh();
  if (!mesh) {
    fprintf(stderr, "Unble to allocate mesh\n");
    return NULL;
  }

  // Read mesh
  if (mesh_name) {
    // Read mesh from file
    if (!mesh->ReadMesh(mesh_name)) {
      fprintf(stderr, "Unble to read mesh from %s\n", mesh_name);
      return NULL;
    }

    // Print statistics
    if (print_verbose) {
      printf("Read mesh from %s ...\n", mesh_name);
      printf("  # Faces = %d\n", mesh->NFaces());
      printf("  # Edges = %d\n", mesh->NEdges());
      printf("  # Vertices = %d\n", mesh->NVertices());
      fflush(stdout);
    }
  }

  // Read function 
  if (function_name) {
    // Read function from file
    if (!mesh->ReadVertexValues(function_name)) {
      fprintf(stderr, "Unble to read function from %s\n", function_name);
      return NULL;
    }

    // Print statistics
    if (print_verbose) {
      printf("Read function from %s ...\n", function_name);
      printf("  # Values = %d\n", mesh->NVertices());
      fflush(stdout);
    }
  }


  // Return mesh
  return mesh;
}



static int
WriteFiles(R3LaplacianMesh *mesh, const char *mesh_name, const char *function_name)
{
  // Write mesh
  if (mesh_name) {
    // Write mesh to file
    if (!mesh->WriteMesh(mesh_name)) {
      fprintf(stderr, "Unble to write mesh to %s\n", mesh_name);
      return 0;
    }

    // Print statistics
    if (print_verbose) {
      printf("Wrote mesh to %s ...\n", mesh_name);
      printf("  # Faces = %d\n", mesh->NFaces());
      printf("  # Edges = %d\n", mesh->NEdges());
      printf("  # Vertices = %d\n", mesh->NVertices());
      fflush(stdout);
    }
  }

  // Write function 
  if (function_name) {
    // Write function to file
    if (!mesh->WriteVertexValues(function_name)) {
      fprintf(stderr, "Unble to write function to %s\n", function_name);
      return 0;
    }

    // Print statistics
    if (print_verbose) {
      printf("Wrote function to %s ...\n", function_name);
      printf("  # Values = %d\n", mesh->NVertices());
      fflush(stdout);
    }
  }

  // Return success
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Viewpoint initialization function
////////////////////////////////////////////////////////////////////////

static R3Viewer *
CreateBirdsEyeViewer(R3LaplacianMesh *mesh)
{
  // Setup camera view looking down the Z axis
  R3Box bbox = mesh->BBox();
  assert(!bbox.IsEmpty());
  center = bbox.Centroid();
  RNLength r = bbox.DiagonalRadius();
  assert((r > 0.0) && RNIsFinite(r));
  static R3Vector towards = R3Vector(-0.57735, -0.57735, -0.57735);
  static R3Vector up = R3Vector(-0.57735, 0.57735, 0.5773);
  R3Point eye = center - towards * (2.5 * r);
  R3Camera camera(eye, towards, up, 0.4, 0.4, 0.01 * r, 100.0 * r);
  R2Viewport viewport(0, 0, GLUTwindow_width, GLUTwindow_height);
  return new R3Viewer(camera, viewport);
}



////////////////////////////////////////////////////////////////////////
// Pick function
////////////////////////////////////////////////////////////////////////

static R3MeshVertex *
PickClosestVertex(R3LaplacianMesh *mesh, int x, int y) 
{
  // Check cursor position
  if ((x < 0) || (GLUTwindow_width <= x) || (y < 0) || (GLUTwindow_height <= y)) {
    printf("Pick point outside viewport\n");
    return NULL;
  }

  // Allocate select buffer
  const int SELECT_BUFFER_SIZE = 1024;
  GLuint select_buffer[SELECT_BUFFER_SIZE];
  GLint select_buffer_hits;

  // Initialize select buffer
  glSelectBuffer(SELECT_BUFFER_SIZE, select_buffer);
  glRenderMode(GL_SELECT);
  glInitNames();
  glPushName(0);

  // Draw faces with pick names into selection buffer
  GLint viewport[4];
  glViewport(0, 0, GLUTwindow_width, GLUTwindow_height);
  glGetIntegerv(GL_VIEWPORT, viewport);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPickMatrix((GLdouble) x, (GLdouble) y, 1, 1, viewport);
  viewer->Camera().Load(TRUE);
  glMatrixMode(GL_MODELVIEW);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  for (int i = 0; i < mesh->NFaces(); i++) {
    R3MeshFace *face = mesh->Face(i);
    glLoadName(i + 1);
    mesh->DrawFace(face);
  }
  glFlush();
  select_buffer_hits = glRenderMode(GL_RENDER);

  // Process select buffer to find front-most hit
  GLuint hit = 0;
  GLuint hit_z = 0xFFFFFFFF;
  GLuint *bufp = select_buffer;
  GLuint numnames, z1, z2;
  for (int i = 0; i < select_buffer_hits; i++) {
    numnames = *bufp++;
    z1 = *bufp++;
    z2 = *bufp++;
    while (numnames--) {
      // This seems wrong?
      if (z1 < hit_z) {
        hit = *bufp;
        hit_z = z1/2 + z2/2;
      }
      bufp++;
    }
  }

  // Check if hit anything
  if (hit <= 0) return NULL;

  // Find hit face
  hit--; // subtract the one added to avoid zero
  if (hit < 0) return NULL;
  if (hit >= (GLuint) mesh->NFaces()) return NULL;
  R3MeshFace *face = mesh->Face(hit);

  // Find hit point
  GLdouble p[3];
  GLdouble modelview_matrix[16];
  GLdouble projection_matrix[16];
  glGetDoublev(GL_MODELVIEW_MATRIX, modelview_matrix);
  glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
  GLdouble z = (GLdouble) hit_z / (GLdouble) 0xFFFFFFFF;
  gluUnProject(x, y, z, modelview_matrix, projection_matrix, viewport, &(p[0]), &(p[1]), &(p[2]));
  R3Point cursor_world_position(p[0], p[1], p[2]);
  R2Point cursor_screen_position(x, y);

  // Find closest vertex on hit face
  RNLength closest_distance = FLT_MAX;
  R3MeshVertex *closest_vertex = NULL;
  for (int i = 0; i < 3; i++) {
    R3MeshVertex *vertex = mesh->VertexOnFace(face, i);
    const R3Point& vertex_world_position = mesh->VertexPosition(vertex);
    R2Point vertex_screen_position; RNScalar dummy;
    gluProject(vertex_world_position[0], vertex_world_position[1], vertex_world_position[2], 
      modelview_matrix, projection_matrix, viewport, 
      &(vertex_screen_position[0]), &(vertex_screen_position[1]), &dummy);
    RNLength distance = R2Distance(vertex_screen_position, cursor_screen_position);
    if (distance < closest_distance) {
      closest_vertex = vertex;
      closest_distance = distance;
    }
  }

  // Return closest vertex
  return closest_vertex;
}



////////////////////////////////////////////////////////////////////////
// Anchor management functions
////////////////////////////////////////////////////////////////////////

static void 
EmptyAnchors(void)
{
  // Reset current anchor
  current_anchor = NULL;

  // Delete all anchors
  anchors.clear();
}



static void
SelectAnchor(int x, int y)
{
  // Reset current anchor
  current_anchor = NULL;

  // Check if selected anchor position
  R2Point cursor_position(x,y);
  for (unsigned int i = 0; i < anchors.size(); i++) {
    Anchor *anchor = anchors[i];
    R2Point anchor_position = viewer->ViewportPoint(anchor->position);
    if (R2Distance(anchor_position, cursor_position) < 10) {
      current_anchor = anchor;
      return;
    }
  }

  // Pick vertex
  R3MeshVertex *vertex = PickClosestVertex(mesh, x, y);
  if (!vertex) return;

  // Check if vertex is an anchor
  for (unsigned int i = 0; i < anchors.size(); i++) {
    Anchor *anchor = anchors[i];
    if (vertex == mesh->Vertex(anchor->vertex_id)) {
      current_anchor = anchor;
      return;
    }
  }

  // Create/add anchor 
  current_anchor = new Anchor();
  current_anchor->vertex_id = mesh->VertexID(vertex);
  current_anchor->position = mesh->VertexPosition(vertex);
  anchors.push_back(current_anchor);
}



static void
DeleteAnchor(int x, int y)
{
  // Reset current anchor
  current_anchor = NULL;

  // Pick vertex
  R3MeshVertex *vertex = PickClosestVertex(mesh, x, y);
  if (!vertex) return;

  // Check if vertex is  an anchor
  int anchor_index = -1;
  for (unsigned int i = 0; i < anchors.size(); i++) {
    Anchor *anchor = anchors[i];
    if (vertex == mesh->Vertex(anchor->vertex_id)) {
      anchor_index = i;
      break;
    }
  }

  // Check if vertex is an anchor
  if (anchor_index == -1) {
    if (print_verbose) printf("Vertex is not an anchor\n");
    return;
  }

  // Remove anchor
  anchors.erase(anchors.begin()+anchor_index);
}


static void
RepositionCurrentAnchor(int x, int y)
{
  // Check if there is a current anchor
  if (!current_anchor) return;

  // Compute new position
  R3MeshVertex *current_anchor_vertex = mesh->Vertex(current_anchor->vertex_id);
  R3Point position = current_anchor->position;
  R3Vector normal = viewer->Camera().Towards();
  normal.Cross(mesh->VertexNormal(current_anchor_vertex));
  normal.Cross(mesh->VertexNormal(current_anchor_vertex));
  if (normal.IsZero()) return;
  R3Plane plane(position, normal);
  R3Ray ray = viewer->WorldRay(x, y);
  if (!R3Intersects(ray, plane, &position)) return;
  current_anchor->position = position;
}



static void
UpdateVertexPositions(void)
{
  // Check if any anchors
  if (anchors.empty()) return;

  // Print message
  if (print_verbose) {
    printf("Updating vertex positions ...\n"); 
    fflush(stdout);
  }

  // Build vector of anchor vertices and set their positions
  vector<R3MeshVertex *> anchor_vertices;
  for (unsigned int i = 0; i < anchors.size(); i++) {
    R3MeshVertex *vertex = mesh->Vertex(anchors[i]->vertex_id);
    mesh->SetVertexPosition(vertex, anchors[i]->position);
    anchor_vertices.push_back(vertex);
  }

  // Update vertex positions
  mesh->UpdateVertexPositions(anchor_vertices);

  // Reset current anchor
  current_anchor = NULL;

  // Print message
  if (print_verbose) {
    printf("  # Anchors = %d\n", anchors.size());
    printf("  # Vertices = %d\n", mesh->NVertices());
    fflush(stdout);
  }
}


static void
UpdateVertexValues(void)
{
  // Check if any anchors
  if (!input_function_name) return;

  // Print message
  if (print_verbose) {
    printf("Updating vertex values ...\n"); 
    fflush(stdout);
  }

  // Update function values
  mesh->UpdateVertexValues(empty_value);

  // Print message
  if (print_verbose) {
    printf("  # Values = %d\n", mesh->NVertices());
    fflush(stdout);
  }
}



////////////////////////////////////////////////////////////////////////
// GLUT callback functions
////////////////////////////////////////////////////////////////////////

void GLUTDrawText(const R3Point& p, const char *s)
{
  // Draw text string s and position p
  glRasterPos3d(p[0], p[1], p[2]);
  while (*s) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *(s++));
}
  


void GLUTStop(void)
{
  // Output mesh file
  WriteFiles(mesh, output_mesh_name, output_function_name);

  // Destroy window 
  glutDestroyWindow(GLUTwindow);

  // Exit
  exit(0);
}



void GLUTRedraw(void)
{
  // Set viewing transformation
  viewer->Camera().Load();

  // Clear window 
  glClearColor(1.0, 1.0, 1.0, 1.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Set backface culling
  if (show_backfacing) glDisable(GL_CULL_FACE);
  else glEnable(GL_CULL_FACE);

  // Set lights
  static GLfloat light1_position[] = { 3.0, 4.0, 5.0, 0.0 };
  glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
  static GLfloat ambient_material[] = { 0.1, 0.1, 0.1, 1.0 };
  static GLfloat diffuse_material[] = { 0.5, 0.5, 1.0, 1.0 };
  static GLfloat specular_material[] = { 0.3, 0.3, 0.3, 1.0};
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient_material); 
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse_material); 
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular_material); 
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 1);

  // Draw faces
  if (show_faces) {
    glEnable(GL_LIGHTING);
    static GLfloat material[] = { 0.7, 0.7, 0.7, 1.0 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
    mesh->DrawFaces();
  }

  // Draw edges
  if (show_edges) {
    glDisable(GL_LIGHTING);
    glColor3f(1, 0, 0);
    glPolygonOffset(-2,2);
    glEnable(GL_POLYGON_OFFSET_LINE);
    mesh->DrawEdges();
    glDisable(GL_POLYGON_OFFSET_LINE);
  }

  // Draw vertices
  if (show_vertices) {
    glEnable(GL_LIGHTING);
    static GLfloat material[] = { 0, 0, 1, 1 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
    RNScalar radius = 0.002 * mesh->BBox().DiagonalLength();
    for (int i = 0; i < mesh->NVertices(); i++) {
      R3MeshVertex *vertex = mesh->Vertex(i);
      R3Point position = mesh->VertexPosition(vertex);
      R3Sphere(position, radius).Draw();
    }
  }

  // Draw anchors
  if (show_anchors) {
    glEnable(GL_LIGHTING);
    static GLfloat material[] = { 1, 1, 0, 1 };
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material); 
    RNScalar radius = 0.004 * mesh->BBox().DiagonalLength();
    for (unsigned int i = 0; i < anchors.size(); i++) {
      Anchor *anchor = anchors[i];
      R3MeshVertex *vertex = mesh->Vertex(anchor->vertex_id);
      R3Sphere(anchor->position, radius).Draw();
      R3Sphere(mesh->VertexPosition(vertex), radius).Draw();
      R3Span(mesh->VertexPosition(vertex), anchor->position).Draw();
    }
  }

  // Draw values
  if (show_values) {
    glDisable(GL_LIGHTING);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(2, 1);
    mesh->DrawVertexValues(empty_value);
    glDisable(GL_POLYGON_OFFSET_FILL);
  }

  // Capture image and exit
  if (output_image_name) {
    if (print_verbose) printf("Creating image %s\n", output_image_name);
    R2Image image(GLUTwindow_width, GLUTwindow_height, 3);
    image.Capture();
    image.Write(output_image_name);
    output_image_name = NULL;
  }

  // Swap buffers 
  glutSwapBuffers();

  // Exit if batch
  if (batch) GLUTStop();
}    



void GLUTResize(int w, int h)
{
  // Resize window
  glViewport(0, 0, w, h);

  // Resize viewer viewport
  viewer->ResizeViewport(0, 0, w, h);

  // Remember window size 
  GLUTwindow_width = w;
  GLUTwindow_height = h;

  // Redraw
  glutPostRedisplay();
}



void GLUTMotion(int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Compute mouse movement
  int dx = x - GLUTmouse[0];
  int dy = y - GLUTmouse[1];
  
  // World in hand navigation 
  if (GLUTmodifiers & GLUT_ACTIVE_CTRL) {
    // Anchor movement
    RepositionCurrentAnchor(x,y);
    glutPostRedisplay();
  }
  else {
    // Camera control
    if (GLUTbutton[0]) viewer->RotateWorld(1.0, center, x, y, dx, dy);
    else if (GLUTbutton[1]) viewer->ScaleWorld(1.0, center, x, y, dx, dy);
    else if (GLUTbutton[2]) viewer->TranslateWorld(1.0, center, x, y, dx, dy);
    if (GLUTbutton[0] || GLUTbutton[1] || GLUTbutton[2]) glutPostRedisplay();
  }

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;
}



void GLUTMouse(int button, int state, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;
  
  // Process mouse click
  if (glutGetModifiers() & GLUT_ACTIVE_CTRL) {
    if (button == GLUT_LEFT_BUTTON) {
      if (state == GLUT_DOWN) {
        SelectAnchor(x, y);
      }
    }
    else if (button == GLUT_RIGHT_BUTTON) {
      if (state == GLUT_DOWN) {
        DeleteAnchor(x, y);
      }
    }
  }

  // Remember button state 
  int b = (button == GLUT_LEFT_BUTTON) ? 0 : ((button == GLUT_MIDDLE_BUTTON) ? 1 : 2);
  GLUTbutton[b] = (state == GLUT_DOWN) ? 1 : 0;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

   // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Redraw
  glutPostRedisplay();
}



void GLUTSpecial(int key, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Process keyboard button event 

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Redraw
  glutPostRedisplay();
}



void GLUTKeyboard(unsigned char key, int x, int y)
{
  // Invert y coordinate
  y = GLUTwindow_height - y;

  // Process keyboard button event 
  switch (key) {
  case '~': {
    // Dump screen shot to file iX.jpg
    static char buffer[64];
    static int image_count = 1;
    sprintf(buffer, "i%d.jpg", image_count++);
    output_image_name = buffer;
    break; }

  case 'A':
  case 'a':
    show_anchors = !show_anchors;
    break;

  case 'B':
  case 'b':
    show_backfacing = !show_backfacing;
    break;

  case 'C': 
  case 'c': {
    // Update viewing center
    R3MeshVertex *vertex = PickClosestVertex(mesh, x, y);
    if (vertex) center = mesh->VertexPosition(vertex);
    break; }

  case 'E':
  case 'e':
    show_edges = !show_edges;
    break;
 
  case 'F':
  case 'f':
    show_faces = !show_faces;
    break;

  case 'Q':
  case 'q':
    // Quit
    GLUTStop();
    break;

  case 'V':
  case 'v':
    show_vertices = !show_vertices;
    break;

  case 27: // ESCAPE
    EmptyAnchors();
    break;

  case ' ':
    // Update after drag anchors
    UpdateVertexPositions();
    UpdateVertexValues();
    break;
  }

  // Remember mouse position 
  GLUTmouse[0] = x;
  GLUTmouse[1] = y;

  // Remember modifiers 
  GLUTmodifiers = glutGetModifiers();

  // Redraw
  glutPostRedisplay();  
}



void GLUTRun(int *argc, char **argv)
{
  // Open window 
  glutInit(argc, argv);
  glutInitWindowPosition(100, 100);
  glutInitWindowSize(GLUTwindow_width, GLUTwindow_height);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH); // | GLUT_STENCIL
  GLUTwindow = glutCreateWindow("OpenGL Viewer");

  // Initialize background color 
  glClearColor(200.0/255.0, 200.0/255.0, 200.0/255.0, 1.0);

  // Initialize lighting
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  static GLfloat lmodel_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  glEnable(GL_NORMALIZE);
  glEnable(GL_LIGHTING); 

  // Initialize headlight
  static GLfloat light0_diffuse[] = { 0.5, 0.5, 0.5, 1.0 };
  static GLfloat light0_position[] = { 0.0, 0.0, 1.0, 0.0 };
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
  glEnable(GL_LIGHT0);

  // Initialize graphics modes 
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);

  // Initialize GLUT callback functions 
  glutDisplayFunc(GLUTRedraw);
  glutReshapeFunc(GLUTResize);
  glutKeyboardFunc(GLUTKeyboard);
  glutSpecialFunc(GLUTSpecial);
  glutMouseFunc(GLUTMouse);
  glutMotionFunc(GLUTMotion);

  // Run main loop -- never returns 
  glutMainLoop();
}



////////////////////////////////////////////////////////////////////////
// Argument parsing function
////////////////////////////////////////////////////////////////////////

static int 
ParseArgs(int argc, char **argv)
{
  // Parse arguments
  argc--; argv++;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if (!strcmp(*argv, "-v")) {
        // Print helpful messages
        print_verbose = 1;
      }
      else if (!strcmp(*argv, "-batch")) {
        // Do not run interactive interface
        batch = 1;
      }
      else if (!strcmp(*argv, "-input_mesh")) { 
        // Name of input mesh file
        argc--; argv++; input_mesh_name = *argv; 
      }
      else if (!strcmp(*argv, "-output_mesh")) { 
        // Name of output mesh file
        argc--; argv++; output_mesh_name = *argv; 
      }
      else if (!strcmp(*argv, "-input_function")) { 
        // Name of input file with function value for every vertex
        argc--; argv++; input_function_name = *argv; 
      }
      else if (!strcmp(*argv, "-output_function")) { 
        // Name of output file with function value for every vertex
        argc--; argv++; output_function_name = *argv; 
      }
      else if (!strcmp(*argv, "-output_image")) { 
        // Name of image to create with screen dump of initial view
        argc--; argv++; output_image_name = *argv; 
      }
      else if (!strcmp(*argv, "-empty_value")) { 
        // Value marking vertices to fill-in when update vertex values
        argc--; argv++; empty_value = atof(*argv); 
      }
      else if (!strcmp(*argv, "-anchor")) {
        // Inserts an anchor -- can be specified many times on command line
        int vertex_id;
        double x, y, z;
        argc--; argv++; vertex_id = atoi(*argv); 
        argc--; argv++; x = atof(*argv); 
        argc--; argv++; y = atof(*argv); 
        argc--; argv++; z = atof(*argv); 
        Anchor *anchor = new Anchor();
        anchor->vertex_id = vertex_id;
        anchor->position = R3Point(x,y,z);
        anchors.push_back(anchor);
      }
      else { 
        // Invalid argument
        fprintf(stderr, "Invalid program argument: %s", *argv); 
        exit(1); 
      }
      argv++; argc--;
    }
    else {
      // Can provide input mesh name as argument by itself
      if (!input_mesh_name) input_mesh_name = *argv;
      else { fprintf(stderr, "Invalid program argument: %s", *argv); exit(1); }
      argv++; argc--;
    }
  }

  // Check filenames
  if (!input_mesh_name) {
    fprintf(stderr, "Usage: lap input_mesh_file <options>\n");
    return 0;
  }

  // Return OK status 
  return 1;
}



////////////////////////////////////////////////////////////////////////
// Main
////////////////////////////////////////////////////////////////////////

int 
main(int argc, char **argv)
{
  // Check number of arguments
  if (!ParseArgs(argc, argv)) exit(1);

  // Read input mesh
  mesh = ReadFiles(input_mesh_name, input_function_name);
  if (!mesh) exit(-1);

  // Update vertex positions
  UpdateVertexPositions();
    
  // Update function
  UpdateVertexValues();
    
  // Check if should run interface 
  if (!batch || output_image_name) {
    // Create viewer
    viewer = CreateBirdsEyeViewer(mesh);
    if (!viewer) exit(-1);

    // Run interactive interface (never returns)
    GLUTRun(&argc, argv);
  }
  else {
    // Just write files
    WriteFiles(mesh, output_mesh_name, output_function_name);
  }

  // Return success 
  return 0;
}









